/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(NETSCAPE_PLUGIN_API)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "MessageNames.h"
#include "NPIdentifierData.h"
#include "NPObjectMessageReceiverMessagesReplies.h"
#include "NPVariantData.h"
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>

namespace WebKit {
class NPIdentifierData;
class NPVariantData;
}

namespace Messages {
namespace NPObjectMessageReceiver {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::NPObjectMessageReceiver;
}

class Deallocate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_Deallocate; }
    static constexpr bool isSync = true;

    using DelayedReply = DeallocateDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HasMethod {
public:
    using Arguments = std::tuple<const WebKit::NPIdentifierData&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_HasMethod; }
    static constexpr bool isSync = true;

    using DelayedReply = HasMethodDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit HasMethod(const WebKit::NPIdentifierData& methodName)
        : m_arguments(methodName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Invoke {
public:
    using Arguments = std::tuple<const WebKit::NPIdentifierData&, const Vector<WebKit::NPVariantData>&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_Invoke; }
    static constexpr bool isSync = true;

    using DelayedReply = InvokeDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue, const WebKit::NPVariantData& resultData);
    using Reply = std::tuple<bool&, WebKit::NPVariantData&>;
    using ReplyArguments = std::tuple<bool, WebKit::NPVariantData>;
    Invoke(const WebKit::NPIdentifierData& methodName, const Vector<WebKit::NPVariantData>& argumentsData)
        : m_arguments(methodName, argumentsData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class InvokeDefault {
public:
    using Arguments = std::tuple<const Vector<WebKit::NPVariantData>&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_InvokeDefault; }
    static constexpr bool isSync = true;

    using DelayedReply = InvokeDefaultDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue, const WebKit::NPVariantData& resultData);
    using Reply = std::tuple<bool&, WebKit::NPVariantData&>;
    using ReplyArguments = std::tuple<bool, WebKit::NPVariantData>;
    explicit InvokeDefault(const Vector<WebKit::NPVariantData>& argumentsData)
        : m_arguments(argumentsData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HasProperty {
public:
    using Arguments = std::tuple<const WebKit::NPIdentifierData&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_HasProperty; }
    static constexpr bool isSync = true;

    using DelayedReply = HasPropertyDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit HasProperty(const WebKit::NPIdentifierData& propertyName)
        : m_arguments(propertyName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetProperty {
public:
    using Arguments = std::tuple<const WebKit::NPIdentifierData&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_GetProperty; }
    static constexpr bool isSync = true;

    using DelayedReply = GetPropertyDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue, const WebKit::NPVariantData& resultData);
    using Reply = std::tuple<bool&, WebKit::NPVariantData&>;
    using ReplyArguments = std::tuple<bool, WebKit::NPVariantData>;
    explicit GetProperty(const WebKit::NPIdentifierData& propertyName)
        : m_arguments(propertyName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetProperty {
public:
    using Arguments = std::tuple<const WebKit::NPIdentifierData&, const WebKit::NPVariantData&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_SetProperty; }
    static constexpr bool isSync = true;

    using DelayedReply = SetPropertyDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    SetProperty(const WebKit::NPIdentifierData& propertyName, const WebKit::NPVariantData& propertyValueData)
        : m_arguments(propertyName, propertyValueData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveProperty {
public:
    using Arguments = std::tuple<const WebKit::NPIdentifierData&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_RemoveProperty; }
    static constexpr bool isSync = true;

    using DelayedReply = RemovePropertyDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit RemoveProperty(const WebKit::NPIdentifierData& propertyName)
        : m_arguments(propertyName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Enumerate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_Enumerate; }
    static constexpr bool isSync = true;

    using DelayedReply = EnumerateDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue, const Vector<WebKit::NPIdentifierData>& identifiersData);
    using Reply = std::tuple<bool&, Vector<WebKit::NPIdentifierData>&>;
    using ReplyArguments = std::tuple<bool, Vector<WebKit::NPIdentifierData>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Construct {
public:
    using Arguments = std::tuple<const Vector<WebKit::NPVariantData>&>;

    static IPC::MessageName name() { return IPC::MessageName::NPObjectMessageReceiver_Construct; }
    static constexpr bool isSync = true;

    using DelayedReply = ConstructDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue, const WebKit::NPVariantData& resultData);
    using Reply = std::tuple<bool&, WebKit::NPVariantData&>;
    using ReplyArguments = std::tuple<bool, WebKit::NPVariantData>;
    explicit Construct(const Vector<WebKit::NPVariantData>& argumentsData)
        : m_arguments(argumentsData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace NPObjectMessageReceiver
} // namespace Messages

#endif // ENABLE(NETSCAPE_PLUGIN_API)
