/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(NETSCAPE_PLUGIN_API)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "DataReference.h"
#include "MessageNames.h"
#include "PluginControllerProxyMessagesReplies.h"
#include "ShareableBitmap.h"
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class AffineTransform;
class IntRect;
class IntSize;
}

namespace WebKit {
class WebKeyboardEvent;
class WebMouseEvent;
class WebWheelEvent;
}

namespace Messages {
namespace PluginControllerProxy {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::PluginControllerProxy;
}

class GeometryDidChange {
public:
    using Arguments = std::tuple<const WebCore::IntSize&, const WebCore::IntRect&, const WebCore::AffineTransform&, float, const WebKit::ShareableBitmap::Handle&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_GeometryDidChange; }
    static constexpr bool isSync = false;

    GeometryDidChange(const WebCore::IntSize& pluginSize, const WebCore::IntRect& clipRect, const WebCore::AffineTransform& pluginToRootViewTransform, float scaleFactor, const WebKit::ShareableBitmap::Handle& backingStoreHandle)
        : m_arguments(pluginSize, clipRect, pluginToRootViewTransform, scaleFactor, backingStoreHandle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class VisibilityDidChange {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_VisibilityDidChange; }
    static constexpr bool isSync = false;

    explicit VisibilityDidChange(bool isVisible)
        : m_arguments(isVisible)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FrameDidFinishLoading {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_FrameDidFinishLoading; }
    static constexpr bool isSync = false;

    explicit FrameDidFinishLoading(uint64_t requestID)
        : m_arguments(requestID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FrameDidFail {
public:
    using Arguments = std::tuple<uint64_t, bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_FrameDidFail; }
    static constexpr bool isSync = false;

    FrameDidFail(uint64_t requestID, bool wasCancelled)
        : m_arguments(requestID, wasCancelled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidEvaluateJavaScript {
public:
    using Arguments = std::tuple<uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_DidEvaluateJavaScript; }
    static constexpr bool isSync = false;

    DidEvaluateJavaScript(uint64_t requestID, const String& result)
        : m_arguments(requestID, result)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StreamWillSendRequest {
public:
    using Arguments = std::tuple<uint64_t, const String&, const String&, uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_StreamWillSendRequest; }
    static constexpr bool isSync = false;

    StreamWillSendRequest(uint64_t streamID, const String& requestURLString, const String& redirectResponseURLString, uint32_t redirectResponseStatusCode)
        : m_arguments(streamID, requestURLString, redirectResponseURLString, redirectResponseStatusCode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StreamDidReceiveResponse {
public:
    using Arguments = std::tuple<uint64_t, const String&, uint32_t, uint32_t, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_StreamDidReceiveResponse; }
    static constexpr bool isSync = false;

    StreamDidReceiveResponse(uint64_t streamID, const String& responseURLString, uint32_t streamLength, uint32_t lastModifiedTime, const String& mimeType, const String& headers)
        : m_arguments(streamID, responseURLString, streamLength, lastModifiedTime, mimeType, headers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StreamDidReceiveData {
public:
    using Arguments = std::tuple<uint64_t, const IPC::DataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_StreamDidReceiveData; }
    static constexpr bool isSync = false;

    StreamDidReceiveData(uint64_t streamID, const IPC::DataReference& data)
        : m_arguments(streamID, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StreamDidFinishLoading {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_StreamDidFinishLoading; }
    static constexpr bool isSync = false;

    explicit StreamDidFinishLoading(uint64_t streamID)
        : m_arguments(streamID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StreamDidFail {
public:
    using Arguments = std::tuple<uint64_t, bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_StreamDidFail; }
    static constexpr bool isSync = false;

    StreamDidFail(uint64_t streamID, bool wasCancelled)
        : m_arguments(streamID, wasCancelled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ManualStreamDidReceiveResponse {
public:
    using Arguments = std::tuple<const String&, uint32_t, uint32_t, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_ManualStreamDidReceiveResponse; }
    static constexpr bool isSync = false;

    ManualStreamDidReceiveResponse(const String& responseURLString, uint32_t streamLength, uint32_t lastModifiedTime, const String& mimeType, const String& headers)
        : m_arguments(responseURLString, streamLength, lastModifiedTime, mimeType, headers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ManualStreamDidReceiveData {
public:
    using Arguments = std::tuple<const IPC::DataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_ManualStreamDidReceiveData; }
    static constexpr bool isSync = false;

    explicit ManualStreamDidReceiveData(const IPC::DataReference& data)
        : m_arguments(data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ManualStreamDidFinishLoading {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_ManualStreamDidFinishLoading; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ManualStreamDidFail {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_ManualStreamDidFail; }
    static constexpr bool isSync = false;

    explicit ManualStreamDidFail(bool wasCancelled)
        : m_arguments(wasCancelled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandleMouseEvent {
public:
    using Arguments = std::tuple<const WebKit::WebMouseEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandleMouseEvent; }
    static constexpr bool isSync = false;

    explicit HandleMouseEvent(const WebKit::WebMouseEvent& mouseEvent)
        : m_arguments(mouseEvent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandleWheelEvent {
public:
    using Arguments = std::tuple<const WebKit::WebWheelEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandleWheelEvent; }
    static constexpr bool isSync = true;

    using DelayedReply = HandleWheelEventDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit HandleWheelEvent(const WebKit::WebWheelEvent& wheelEvent)
        : m_arguments(wheelEvent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandleMouseEnterEvent {
public:
    using Arguments = std::tuple<const WebKit::WebMouseEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandleMouseEnterEvent; }
    static constexpr bool isSync = true;

    using DelayedReply = HandleMouseEnterEventDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit HandleMouseEnterEvent(const WebKit::WebMouseEvent& mouseEvent)
        : m_arguments(mouseEvent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandleMouseLeaveEvent {
public:
    using Arguments = std::tuple<const WebKit::WebMouseEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandleMouseLeaveEvent; }
    static constexpr bool isSync = true;

    using DelayedReply = HandleMouseLeaveEventDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit HandleMouseLeaveEvent(const WebKit::WebMouseEvent& mouseEvent)
        : m_arguments(mouseEvent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandleKeyboardEvent {
public:
    using Arguments = std::tuple<const WebKit::WebKeyboardEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandleKeyboardEvent; }
    static constexpr bool isSync = true;

    using DelayedReply = HandleKeyboardEventDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit HandleKeyboardEvent(const WebKit::WebKeyboardEvent& keyboardEvent)
        : m_arguments(keyboardEvent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandleEditingCommand {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandleEditingCommand; }
    static constexpr bool isSync = true;

    using DelayedReply = HandleEditingCommandDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    HandleEditingCommand(const String& commandName, const String& argument)
        : m_arguments(commandName, argument)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IsEditingCommandEnabled {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_IsEditingCommandEnabled; }
    static constexpr bool isSync = true;

    using DelayedReply = IsEditingCommandEnabledDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool enabled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit IsEditingCommandEnabled(const String& commandName)
        : m_arguments(commandName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HandlesPageScaleFactor {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_HandlesPageScaleFactor; }
    static constexpr bool isSync = true;

    using DelayedReply = HandlesPageScaleFactorDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool enabled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RequiresUnifiedScaleFactor {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_RequiresUnifiedScaleFactor; }
    static constexpr bool isSync = true;

    using DelayedReply = RequiresUnifiedScaleFactorDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool required);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetFocus {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_SetFocus; }
    static constexpr bool isSync = false;

    explicit SetFocus(bool isFocused)
        : m_arguments(isFocused)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidUpdate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_DidUpdate; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PaintEntirePlugin {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_PaintEntirePlugin; }
    static constexpr bool isSync = true;

    using DelayedReply = PaintEntirePluginDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetPluginScriptableNPObject {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_GetPluginScriptableNPObject; }
    static constexpr bool isSync = true;

    using DelayedReply = GetPluginScriptableNPObjectDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, uint64_t pluginScriptableNPObjectID);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class WindowFocusChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_WindowFocusChanged; }
    static constexpr bool isSync = false;

    explicit WindowFocusChanged(bool hasFocus)
        : m_arguments(hasFocus)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class WindowVisibilityChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_WindowVisibilityChanged; }
    static constexpr bool isSync = false;

    explicit WindowVisibilityChanged(bool isVisible)
        : m_arguments(isVisible)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class SendComplexTextInput {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_SendComplexTextInput; }
    static constexpr bool isSync = false;

    explicit SendComplexTextInput(const String& textInput)
        : m_arguments(textInput)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class WindowAndViewFramesChanged {
public:
    using Arguments = std::tuple<const WebCore::IntRect&, const WebCore::IntRect&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_WindowAndViewFramesChanged; }
    static constexpr bool isSync = false;

    WindowAndViewFramesChanged(const WebCore::IntRect& windowFrameInScreenCoordinates, const WebCore::IntRect& viewFrameInWindowCoordinates)
        : m_arguments(windowFrameInScreenCoordinates, viewFrameInWindowCoordinates)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetLayerHostingMode {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_SetLayerHostingMode; }
    static constexpr bool isSync = false;

    explicit SetLayerHostingMode(uint32_t layerHostingMode)
        : m_arguments(layerHostingMode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SupportsSnapshotting {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_SupportsSnapshotting; }
    static constexpr bool isSync = true;

    using DelayedReply = SupportsSnapshottingDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool isSupported);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Snapshot {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_Snapshot; }
    static constexpr bool isSync = true;

    using DelayedReply = SnapshotDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const WebKit::ShareableBitmap::Handle& backingStoreHandle);
    using Reply = std::tuple<WebKit::ShareableBitmap::Handle&>;
    using ReplyArguments = std::tuple<WebKit::ShareableBitmap::Handle>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StorageBlockingStateChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_StorageBlockingStateChanged; }
    static constexpr bool isSync = false;

    explicit StorageBlockingStateChanged(bool storageBlockingEnabled)
        : m_arguments(storageBlockingEnabled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PrivateBrowsingStateChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_PrivateBrowsingStateChanged; }
    static constexpr bool isSync = false;

    explicit PrivateBrowsingStateChanged(bool isPrivateBrowsingEnabled)
        : m_arguments(isPrivateBrowsingEnabled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetFormValue {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_GetFormValue; }
    static constexpr bool isSync = true;

    using DelayedReply = GetFormValueDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool returnValue, const String& formValue);
    using Reply = std::tuple<bool&, String&>;
    using ReplyArguments = std::tuple<bool, String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class MutedStateChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginControllerProxy_MutedStateChanged; }
    static constexpr bool isSync = false;

    explicit MutedStateChanged(bool muted)
        : m_arguments(muted)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace PluginControllerProxy
} // namespace Messages

#endif // ENABLE(NETSCAPE_PLUGIN_API)
