/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(NETSCAPE_PLUGIN_API)

#include "ArgumentCoders.h"
#include "Connection.h"
#include "MessageNames.h"
#include "PluginProcessProxyMessagesReplies.h"
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace IPC {
class Attachment;
}

namespace Messages {
namespace PluginProcessProxy {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::PluginProcessProxy;
}

class DidCreateWebProcessConnection {
public:
    using Arguments = std::tuple<const IPC::Attachment&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_DidCreateWebProcessConnection; }
    static constexpr bool isSync = false;

    DidCreateWebProcessConnection(const IPC::Attachment& connectionIdentifier, bool supportsAsynchronousPluginInitialization)
        : m_arguments(connectionIdentifier, supportsAsynchronousPluginInitialization)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidGetSitesWithData {
public:
    using Arguments = std::tuple<const Vector<String>&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_DidGetSitesWithData; }
    static constexpr bool isSync = false;

    DidGetSitesWithData(const Vector<String>& sites, uint64_t callbackID)
        : m_arguments(sites, callbackID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDeleteWebsiteData {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_DidDeleteWebsiteData; }
    static constexpr bool isSync = false;

    explicit DidDeleteWebsiteData(uint64_t callbackID)
        : m_arguments(callbackID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDeleteWebsiteDataForHostNames {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_DidDeleteWebsiteDataForHostNames; }
    static constexpr bool isSync = false;

    explicit DidDeleteWebsiteDataForHostNames(uint64_t callbackID)
        : m_arguments(callbackID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class SetModalWindowIsShowing {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_SetModalWindowIsShowing; }
    static constexpr bool isSync = false;

    explicit SetModalWindowIsShowing(bool modalWindowIsShowing)
        : m_arguments(modalWindowIsShowing)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetFullscreenWindowIsShowing {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_SetFullscreenWindowIsShowing; }
    static constexpr bool isSync = false;

    explicit SetFullscreenWindowIsShowing(bool fullscreenWindowIsShowing)
        : m_arguments(fullscreenWindowIsShowing)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class LaunchProcess {
public:
    using Arguments = std::tuple<const String&, const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_LaunchProcess; }
    static constexpr bool isSync = true;

    using DelayedReply = LaunchProcessDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    LaunchProcess(const String& launchPath, const Vector<String>& arguments)
        : m_arguments(launchPath, arguments)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class LaunchApplicationAtURL {
public:
    using Arguments = std::tuple<const String&, const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_LaunchApplicationAtURL; }
    static constexpr bool isSync = true;

    using DelayedReply = LaunchApplicationAtURLDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    LaunchApplicationAtURL(const String& url, const Vector<String>& arguments)
        : m_arguments(url, arguments)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class OpenURL {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_OpenURL; }
    static constexpr bool isSync = true;

    using DelayedReply = OpenURLDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool result, int32_t status, const String& launchedURLString);
    using Reply = std::tuple<bool&, int32_t&, String&>;
    using ReplyArguments = std::tuple<bool, int32_t, String>;
    explicit OpenURL(const String& urlString)
        : m_arguments(urlString)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class OpenFile {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::PluginProcessProxy_OpenFile; }
    static constexpr bool isSync = true;

    using DelayedReply = OpenFileDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit OpenFile(const String& fullPath)
        : m_arguments(fullPath)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace PluginProcessProxy
} // namespace Messages

#endif // ENABLE(NETSCAPE_PLUGIN_API)
